#******************************************************************************
#*** Execute Programs in a TCL window *****************************************
#******************************************************************************

proc ExecProg {command fname} {

#***
#*** Execute command. Command may include the symbol $fname, which
#*** will be replaced by filename. $fnamewoext will be replaced by
#*** filename without extension 
#*** If fname == "", use mainname or editfname without extension
#*** Before executing, go to directory of fname
#***

        global editfname editwin
        global mainflag mainname
        global readchild readchierr writechild
        global childpid 
        global boldl_font
        global chisaveflag chireloadflag

#*** Is there another programm running ?

        if {[childstat]} {
                DisplayInfo "Sorry:\nThere is another program running\n\nKill other program first" "$boldl_font"
                return
        }
                
#*** Determine fname as explained above
        
        if {[cequal "" $fname]} {
                if {[info exists editfname]} {
                        set fname $editfname
                }
                if {$mainflag} {set fname $mainname}
        }            

#*** Does filename make sense ?

        if {[cequal "" $fname]} {
                if {[info exists editwin] && [winfo exists $editwin]} {
			DisplayInfo "Please save your file before executing this command.\n I don't know the name of your file yet" $boldl_font
		} else {
                	DisplayInfo "No File has been specified so far.\nCannot execute command!\nPlease load a file in the editor or specify a main file" $boldl_font
		}
                return
        }

#*** filename without extension --> fnamewoext

        set fnamewoext "$fname"
        set pos  [expr [string last "." "$fname"] - 1]
        set pos1 [expr [string last "/" "$fname"] - 1]; #*** Take care of "." in directory name 

        if {$pos >= 0 && $pos > $pos1} {
                set fnamewoext [crange "$fname" 0 "$pos" ]
        }
        
#*** Replace macros :save :reload :fname :fnamewoext
        
        set chisaveflag false
        set chireloadflag false

        while {1} {
                set pos [string first ":" "$command"]
                if {$pos==-1} {break}

                set sstr [expr {$pos ? [crange "$command" 0 "$pos-1" ] : ""}]
                set estr [crange "$command" "$pos+1" end]
                set mstr [string tolower [ctoken estr " .,:"]]

                if {[cequal "$mstr" "fnamewoext"]} {
                        set mstr $fnamewoext
                }
                if {[cequal "$mstr" "fname"]} {
                        set mstr $fname
                }
                if {[cequal "$mstr" "save"]} {
                        set mstr ""
                        set chisaveflag true
                }
                if {[cequal "$mstr" "reload"]} {
                        set mstr ""
                        set chireloadflag true
                }
                set command [format "%s%s%s" "$sstr" "$mstr" "$estr"]
        }
        
#*** Create window if it doesn't exist

        global  exec_win    execwin_name    execwin_geo
        global  childpid

        set     exec_win    $execwin_name
        set     c           $execwin_name.c

        if {[cequal "" [CreateTopWin execwin ""]]} {
                wm withdraw  $exec_win
                wm deiconify $exec_win
        } else {
                wm title    $exec_win "$command"
                wm iconname $exec_win "$command"
          
                frame       $exec_win.but        -relief raised -borderwidth 1 -width 80c -height 10c
                pack        $exec_win.but        -side bottom -fill x

                button      $exec_win.but.kill   -text "Kill Program"  -command {childkill}
                button      $exec_win.but.done   -text "Done"  -command {childkill; destroy $exec_win}
                button      $exec_win.but.clear  -text "Clear" -command {$exec_win.c delete 0.0 end}
                pack        $exec_win.but.kill   $exec_win.but.clear  -side left -anchor s -fill x -padx 3m -pady 1m -ipadx 0.6m
                pack        $exec_win.but.done  -side right -anchor s -fill x -padx 3m -pady 1m -ipadx 0.6m

                scrollbar   $exec_win.ys         -command "$c yview" -relief sunken
                pack        $exec_win.ys         -side right -fill y

                text        $c                   -relief raised -bd 2 -yscrollcommand "$exec_win.ys set"
                pack        $c                   -side left -fill both -expand yes  

                $c tag configure error           -foreground red
                $c tag configure message         -foreground blue 
 
                bind        $c <Any-Return>      { %W insert insert "\n"
                                                   puts $writechild ""
                                                   %W yview -pickplace insert
                                                 }
                bind        $c <Any-KeyPress>    { if {[childstat]} {
                                                        %W insert insert %A
                                                        puts -nonewline $writechild "%A"
                                                        %W yview -pickplace insert
                                                   }            
                                                 }
		bind        $c <1>		 { focus %W}
        }


#*** Save text ?

        global editor
        if {$chisaveflag} {
                if {[cequal "$editor" "XTeX"] && [Qmodflag]} {SaveFile ""}
        }

#*** Compose Command, program-name should go to prog, parameters to command

        eval  set command \"$command\"
        set   progname [lvarpop command 0]

        set   childpath [string range $fname 0 [string last "/" $fname]] 

#*** Create pipes to enable communication with child 

        pipe  readchild   writepar;                     #*** Child --> Parent                   
        fcntl $readchild  NONBLOCK 1
        fcntl $readchild  NOBUF    1

        pipe  readchierr  writeparerr;                  #*** Child stderr --> Parent
        fcntl $readchierr NONBLOCK 1
        fcntl $readchierr NOBUF    1

        pipe  readpar     writechild;                   #*** Parent --> Child
        fcntl $writechild NOBUF    1

#*** Now fork and start program
        
        InsertWithTags $c "Executing: $progname $command\n\n" message

        if {[set childpid [fork]] == 0} {
                dup $readpar stdin;                     #*** Child process, set IO to pipes     
                close $readpar
                dup $writepar stdout
                close $writepar
                dup $writeparerr stderr
                close $writeparerr

                pushd $childpath;                       #*** Change to data file directory
                execl $progname $command                

                puts stderr "*** Sorry: Programm could not be started"
                kill [id process]               
                exit 255
        }

#*** Disable buffering and install driver to update window 5x per second

	focus  $c
        update
        after  200 {childIO $exec_win.c}
}

proc    childkill {} {

#*** Kill the child process

        global childpid
        if {[childstat]} {
                kill 9 $childpid
        }
}


proc    childstat {} {

#*** Get status of child. (1 = child exists, 0 = child does not exist)

        global childpid
        if {![info exists childpid]} {set childpid 0; return 0}

        set fname "/proc/$childpid"
        if {$childpid} {
                if {[file isdirectory $fname]} {return 1}
                set childpid 0
        }
        return 0
}


proc    childIO {W} {

#*** This process runs in background during execution of a unix command
#*** It takes the output from the pipe and fills it in the window. It
#*** restarts itself with a delay of 200ms

        global childpid readchild readchierr
        global chireloadflag
        set chibarline "\n----------------------------------------------------------------------------\n"

        if {![childstat]} {return}

#*** Read Standard-Error from child 

        set pmode [lindex [select [list $readchierr $readchild] {} {} 0] 0]

        if {[string first $readchierr $pmode] >=0 } {
                set line [read $readchierr]
                if {[winfo exists $W]} {
                        InsertWithTags $W "$line" error
                        $W yview -pickplace insert
                }        
        }

#*** Read Standard-Output from child

        if {[string first $readchild $pmode] >=0 } {
                set line [read $readchild] 
                if {[winfo exists $W]} {
                        InsertWithTags $W "$line" ""
                        $W yview -pickplace insert
                }        
        }

#*** Check status of child. If program terminated print message, else restart function after 200ms

        set pmode [wait -nohang $childpid]

        if {$childpid == [lindex $pmode 0]} {
                if {[cequal "EXIT" [lindex $pmode 1]]} {
                        set msg [format "\nProgram terminated, ERC: %d" [lindex $pmode 2]]     
                } else {
                        set msg [format "\n\nProgram was killed by signal: %s" [lindex $pmode 2]]
                }
                if {[winfo exists $W]} {
                        InsertWithTags $W "$msg" message
                        InsertWithTags $W $chibarline ""
                        $W yview -pickplace insert
                }
                global editor
                if {$chireloadflag} {
                        if {[cequal $editor "XTeX"]} { ReLoadFile }
                }
        } else {
                after 200 {childIO $exec_win.c}
        }
}
